<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Command;

use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationMappingConfig;
use Dedi\SyliusMigrationPlugin\Definition\ColumnMappingDefinition;
use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Dedi\SyliusMigrationPlugin\Dto\IdentifierMappingDefinition;
use Dedi\SyliusMigrationPlugin\Dto\MigrationOption;
use Dedi\SyliusMigrationPlugin\Repository\MappingRepository;
use Dedi\SyliusMigrationPlugin\Service\ReaderInterface;
use Doctrine\DBAL\Connection;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

class DebugIdentifierMappingCommand extends Command
{
    private string $logDir;

    private MigrationMappingConfig $mappingRegistry;

    private MappingRepository $mappingRepository;

    private MigrationConfig $migrationConfig;

    private ReaderInterface $reader;

    private Connection $targetConnection;

    public function __construct(
        string $name = null,
        MigrationMappingConfig $mappingRegistry,
        MappingRepository $mappingRepository,
        ReaderInterface $reader,
        MigrationConfig $migrationConfig,
        Connection $targetConnection,
        string $logDir,
    ) {
        parent::__construct($name);
        $this->mappingRegistry = $mappingRegistry;
        $this->mappingRepository = $mappingRepository;
        $this->reader = $reader;
        $this->logDir = $logDir . (new \DateTime())->format('YmdHis') . '_migrations';
        $this->targetConnection = $targetConnection;
        $this->migrationConfig = $migrationConfig;
    }

    protected function configure(): void
    {
        $this
            ->setDescription('Debugs mapping information for migration')
            ->addOption('dump-missing-lines', 'd', InputOption::VALUE_NONE, 'Dumps missing lines into files')
            ->addOption('count-lines', 'c', InputOption::VALUE_NONE, 'Counts lines from source vs lines mapped in destination')
            ->addOption('wrong-mapping', 'w', InputOption::VALUE_NONE, 'Count lines from mapping table that do no exist in destination table')
            ->addOption('suppress-wrong-mapping', 's', InputOption::VALUE_NONE, 'Deletes lines from mapping table that do no exist in destination table')
        ;
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $dumpMissingLines = $input->getOption('dump-missing-lines');
        $countLines = $input->getOption('count-lines');
        $wrongMapping = $input->getOption('wrong-mapping');
        $suppressWrongMapping = $input->getOption('suppress-wrong-mapping');

        $this->doDebugMapping($output);

        $migrationOption = new MigrationOption(true, false, null, null);

        if ($countLines) {
            $this->doCountLines($output, $migrationOption);
        }

        if ($dumpMissingLines) {
            $this->doDumpMissingLines($output, $migrationOption);
        }

        if ($wrongMapping || $suppressWrongMapping) {
            $this->doHandleWrongMapping($output, $suppressWrongMapping);
        }

        $output->writeln('<fg=green>The process finished correctly</fg=green>');

        return Command::SUCCESS;
    }

    private function doDumpMissingLines(OutputInterface $output, MigrationOption $migrationOption): void
    {
        $output->writeln('');
        $output->writeln('');
        $output->writeln('');
        $output->writeln('<info>*************************</info>');
        $output->writeln(sprintf('<info>Dump lines into %s</info>', $this->logDir));
        $output->writeln('<info>*************************</info>');

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                $toExcludeIds = array_map(fn ($identifier) => $identifier['source_id'], $this->mappingRepository->getAll(
                    $identifierMappingDefinition->getSourceTableName(),
                    $identifierMappingDefinition->getSourceColumnName(),
                    $identifierMappingDefinition->getDestTableName(),
                ));
                $linesToImport = $this->reader->getLinesNoInTmpMappingTable($tableDefinition, $migrationOption, $toExcludeIds, -1);
                if (!empty($linesToImport)) {
                    $this->writeLinesInToFile($tableDefinition, $linesToImport);
                }
            }
        }
        $output->writeln(sprintf('<info>Wrote missing lines to files in %s</info>', $this->logDir));
    }

    private function doCountLines(OutputInterface $output, MigrationOption $migrationOption): void
    {
        $output->writeln('');
        $output->writeln('');
        $output->writeln('');
        $output->writeln('<info>*************************</info>');
        $output->writeln('<info>Dump data statistics</info>');
        $output->writeln('<info>*************************</info>');

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            $toImportDataCount = $this->reader->getResultCount($tableDefinition, $migrationOption, -1);

            /** @var IdentifierMappingDefinition $identifierMappingDefinition */
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                $alreadyExistingDataCount = $this->mappingRepository->countIdentifiers(
                    $identifierMappingDefinition->getSourceTableName(),
                    $identifierMappingDefinition->getSourceColumnName(),
                    $identifierMappingDefinition->getDestTableName(),
                );

                if ($toImportDataCount === $alreadyExistingDataCount) {
                    $tag = '<fg=green>';
                    $closingTag = '</>';
                } else {
                    $tag = '<fg=red>';
                    $closingTag = '</>';
                }
                $output->writeln(sprintf(
                    '%sFrom %s to %s (identifier : %s.%s -> %s.%s), import %s row, %s identifier found%s',
                    $tag,
                    $tableDefinition->getSourceTableName() === '' ? 'sql' : $tableDefinition->getSourceTableName(),
                    $tableDefinition->getTargetTableName(),
                    $identifierMappingDefinition->getSourceTableName(),
                    $identifierMappingDefinition->getSourceColumnName(),
                    $identifierMappingDefinition->getDestTableName(),
                    $identifierMappingDefinition->getDestColumnName(),
                    $toImportDataCount,
                    $alreadyExistingDataCount,
                    $closingTag,
                ));
            }
        }
    }

    private function doDebugMapping(OutputInterface $output): void
    {
        $output->writeln('<info>*************************</info>');
        $output->writeln('<info>Dump mapping informations</info>');
        $output->writeln('<info>*************************</info>');

        $data = [];
        $alreadyExistingMappingIdentifiers = [];
        foreach ($this->mappingRepository->getMappingIdentifiers() as $identifier) {
            $alreadyExistingMappingIdentifiers[$identifier['source_table_name'] . '.' . $identifier['source_column_name']] = $identifier;
        }

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            /** @var IdentifierMappingDefinition $identifierMappingDefinition */
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                $data[$identifierMappingDefinition->getSourceTableName() . '.' . $identifierMappingDefinition->getSourceColumnName()] = [];
            }
        }

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            /** @var IdentifierMappingDefinition $identifierMappingDefinition */
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                /** @var ColumnMappingDefinition $columnMappingDefinition */
                foreach ($tableDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
                    $identifierMapping = $columnMappingDefinition->getIdentifierMapping();
                    if (null !== $identifierMapping) {
                        $uid = $identifierMapping['source_table'] . '.' . $identifierMapping['source_column'];

                        if (!array_key_exists($uid, $data)) {
                            $output->writeln('<fg=red>**** ERROR *****</fg=red>');
                            $output->writeln(sprintf('<fg=red>Identifier %s not found, declared on table %s, column %s</fg=red>', $uid, $tableDefinition->getTargetTableName(), $columnMappingDefinition->getTargetColumName()));
                            $output->writeln('<fg=red>****************</fg=red>');

                            continue;
                        }

                        if (null === $identifierMapping['dest_table']) {
                            $data[$uid]['nodest'][] = $tableDefinition->getTargetTableName() . '.' . $columnMappingDefinition->getTargetColumName();
                        } else {
                            if (!array_key_exists($identifierMapping['dest_table'], $data[$uid])) {
                                $data[$uid][$identifierMapping['dest_table']] = [];
                            }
                            $data[$uid][$identifierMapping['dest_table']][] = $tableDefinition->getTargetTableName() . '.' . $columnMappingDefinition->getTargetColumName();
                        }
                    }
                }
            }
        }

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            /** @var IdentifierMappingDefinition $identifierMappingDefinition */
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                $line = $identifierMappingDefinition->getSourceColumnName() !== $identifierMappingDefinition->getSourceTableName() . '_id' ? '[CUSTOM] ' : '[CLASSIC] ';

                $line .= sprintf(
                    '%s => %s.%s => %s.%s',
                    $tableDefinition->getTargetTableName(),
                    $identifierMappingDefinition->getSourceTableName(),
                    $identifierMappingDefinition->getSourceColumnName(),
                    $identifierMappingDefinition->getDestTableName(),
                    $identifierMappingDefinition->getDestColumnName(),
                );
                $output->writeln($line);

                if (!empty($data[$identifierMappingDefinition->getSourceTableName() . '.' . $identifierMappingDefinition->getSourceColumnName()])) {
                    $line = '       [UTILISATIONS] :';
                    $output->writeln($line);

                    foreach ($data[$identifierMappingDefinition->getSourceTableName() . '.' . $identifierMappingDefinition->getSourceColumnName()] as $dest => $columns) {
                        if (empty($columns)) {
                            continue;
                        }
                        $line = '              ' . $dest . ' : ';
                        foreach ($columns as $column) {
                            $line .= $column . ',';
                        }
                        $output->writeln($line);
                    }
                }
            }
        }

        foreach ($alreadyExistingMappingIdentifiers as $uid => $alreadyExistingMappingIdentifier) {
            if (!array_key_exists($uid, $data)) {
                $output->writeln('<fg=red>**** ERROR *****</fg=red>');
                $output->writeln(sprintf('<fg=red>Identifier %s found in mapping table but not in migration</fg=red>', $uid));
                $output->writeln('<fg=red>****************</fg=red>');
            }
        }
    }

    /**
     * @throws \Doctrine\DBAL\Exception
     */
    private function doHandleWrongMapping(OutputInterface $output, mixed $suppressWrongMapping): void
    {
        $output->writeln('');
        $output->writeln('');
        $output->writeln('');
        $output->writeln('<info>**************************</info>');
        $output->writeln('<info>Detecting mapping to clean</info>');
        $output->writeln('<info>**************************</info>');

        /** @var TableMappingDefinition $tableDefinition */
        foreach ($this->mappingRegistry->getTableDefinitions() as $tableDefinition) {
            /** @var IdentifierMappingDefinition $identifierMappingDefinition */
            foreach ($tableDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
                $queryBuilder = $this->targetConnection->createQueryBuilder();
                $queryBuilder
                    ->select('*')
                    ->from($this->migrationConfig->getTempMappingTableName())
                    ->where('dest_table_name = :dest_table_name')
                    ->andWhere('dest_column_name = :dest_column_name')
                    ->andWhere("dest_id NOT IN (SELECT DISTINCT {$identifierMappingDefinition->getDestColumnName()} FROM {$identifierMappingDefinition->getDestTableName()})")
                    ->setParameter('dest_table_name', $identifierMappingDefinition->getDestTableName())
                    ->setParameter('dest_column_name', $identifierMappingDefinition->getDestColumnName())
                ;

                $rowCount = $queryBuilder->execute()->rowCount();
                if (0 === $rowCount) {
                    $tag = '<fg=green>';
                    $closingTag = '</>';
                } else {
                    $tag = '<fg=red>';
                    $closingTag = '</>';
                }

                $output->writeln(sprintf(
                    '%sFrom %s to %s, %s identifiers found that do not exist in source DB%s',
                    $tag,
                    $tableDefinition->getSourceTableName() === '' ? 'sql' : $tableDefinition->getSourceTableName(),
                    $tableDefinition->getTargetTableName(),
                    $rowCount,
                    $closingTag,
                ));

                if ($rowCount > 0 && $suppressWrongMapping) {
                    $deleteQueryBuilder = $this->targetConnection->createQueryBuilder();
                    $deleteQueryBuilder
                        ->delete($this->migrationConfig->getTempMappingTableName())
                        ->where('dest_table_name = :dest_table_name')
                        ->andWhere('dest_column_name = :dest_column_name')
                        ->andWhere("dest_id NOT IN (SELECT DISTINCT {$identifierMappingDefinition->getDestColumnName()} FROM {$identifierMappingDefinition->getDestTableName()})")
                        ->setParameter('dest_table_name', $identifierMappingDefinition->getDestTableName())
                        ->setParameter('dest_column_name', $identifierMappingDefinition->getDestColumnName())
                    ;

                    $deleteQueryBuilder->execute();
                    $output->writeln('<fg=green>Cleaned</>');
                }
            }
        }
    }

    private function writeLinesInToFile(TableMappingDefinition $tableDefinition, array $linesToImport)
    {
        if (!is_dir($this->logDir)) {
            if (!mkdir($concurrentDirectory = $this->logDir, 0755, true) && !is_dir($concurrentDirectory)) {
                throw new \RuntimeException(sprintf('Directory "%s" was not created', $concurrentDirectory));
            }
        }
        $file = fopen(sprintf($this->logDir . '/%s.csv', $tableDefinition->getTargetTableName()), 'wb+');

        if (empty($linesToImport)) {
            return;
        }

        $headers = array_keys($linesToImport[0]);

        fputcsv($file, $headers);
        foreach ($linesToImport as $lineToImport) {
            fputcsv($file, array_values($lineToImport));
        }
        fclose($file);
    }
}
